from grovepi import *
from grove_rgb_lcd import *
import time
import datetime
import threading
import paho.mqtt.client as mqtt
import json
    
# === Variables ===
# Port Assignments
light_port = 1     # A1
pir_port = 4       # D4
button_port = 3    # D2 (FOR SOME REASON THIS BOARD THINKS D3 IS D2)
buzzer_port = 7    # D7

# Thingsboard Connection
client = None
THINGSBOARD_HOST = "thingsboard.cs.cf.ac.uk"
ACCESS_TOKEN = "ad9EfCbasXFLokL3tkC1"

# Notification Preferences
preferences = {"Thingsboard": True,
                 "Visual": True,
                  "Audio": True}

# light variables
light_dark_threshold = 300  # Below this = night
night_mode = False # False = day, True = night
light_dark_counter = 0 # Counts how many times a "night" value is given
light_trigger_delay = 15  # Scans every 0.2 seconds
                        # Meaning: Delay = ScanFrequency * TriggerDelay
                            # Delay being how much time the light sensor has to be in the dark for in seconds
                        # TriggerDelay = Delay/ScanFrequency
                        # 3/0.2 = 15
                        # This means if we want a delay of 3 seconds to use a TriggerDelay value of 15

# motion variables
motion_detected = False
motion_counter = 0
no_motion_counter = 0
last_motion_time = 0

motion_trigger_delay = 7
no_motion_release_delay = 15 # 15 no motion reads to reset
motion_cooldown = 3 #3s between alerts

# button variables
last_button_state = 1  # For state change detection

# === Initilisation Functions ===
def initialise_sensors():
    pinMode(pir_port, "INPUT")
    pinMode(button_port, "INPUT")
    pinMode(buzzer_port, "OUTPUT")
    return

def initilise_thingsboard():
    global client
    try:
        client = mqtt.Client()
        client.username_pw_set(ACCESS_TOKEN)
        client.connect(THINGSBOARD_HOST, 1883, 60)
        client.loop_start()
        client.subscribe("v1/devices/me/attributes")
        client.on_message = handle_attribute_update
        log_event("INFO","Connected to thingsboard")
    except Exception as e:
        log_event("ERROR",f"MQTT Connection failed: {e}")
        return
    
# === Notification Functions ===
def notify(preferences, message = "visitor", duration = 0.2):
    if preferences.get("Thingsboard"):
        notify_thingsboard(message)
    if preferences.get("Visual"):
        notify_visual()
    if preferences.get("Audio"):
        notify_audio(duration)
    return

def notify_visual():
    log_event("NOTIFICATION","LIGHT NOTIFCATION ACTIVATED")
    print("*LIGHTS ACTIVATE*")
    return

def notify_thingsboard(message):
    log_event("NOTIFICATION", f"THINGSBOARD MESSAGE SENT: {message}")
    print(f"[THINGSBOARD] {message}")
    return

def notify_audio(duration):
    log_event("NOTIFICATION", f"BUZZER ACTIVATED FOR {duration}s")
    digitalWrite(buzzer_port, 1)
    time.sleep(duration)
    digitalWrite(buzzer_port, 0)
    return


# === Sensing Functions ===
def detect_button_press():
    global last_button_state
    button = digitalRead(button_port)

    #makes sure button only goes once if held down 
    if button == 0 and last_button_state == 1:
        log_event("EVENT", "Button Pressed")
        set_display("Ringing")
        notify(preferences)
        time.sleep(2)
        set_display("Doorbell System Ready")

    last_button_state = button
    return

def detect_motion():
    global motion_detected, motion_counter, last_motion_time, no_motion_counter
    
    motion = digitalRead(pir_port)
    now = time.time()
    
    if motion == 1:
        motion_counter += 1
        no_motion_counter = 0
        
        if motion_counter >= motion_trigger_delay and not motion_detected:
            if now - last_motion_time >= motion_cooldown:
                motion_detected = True
                last_motion_time = now
                print("MOTION")
    else:
        if motion_detected:
            no_motion_counter += 1
            if no_motion_counter >= no_motion_release_delay:
                motion_detected = False
                motion_counter = 0
                print("MOTION ENDED")
        else:
            motion_counter = 0
            no_motion_counter = 0
    
def detect_light_level():
    global light_dark_counter, light_trigger_delay, night_mode
    light = analogRead(light_port)
    
    if light < light_dark_threshold:
        light_dark_counter += 1
        if light_dark_counter >= light_trigger_delay and not night_mode:
            #log_event("EVENT", "SIGNIFICANT DARKNESS DETECTED")
            adjust_lights()
        else:
            if night_mode:
                #log_event("EVENT", "SIGNIFICANT LIGHT DETECTED")
                light_dark_counter = 0
    else:
        light_dark_counter = 0  # Reset if it's bright

    

# === Visitor Side Functions ===
def set_display(message):
    try:
        if len(message) > 32:
            # Launch scroll in a separate thread
            # Prevents the LCD Display interrupting the main loop
            scroll_thread = threading.Thread(target=scroll_text, args=(message,))
            scroll_thread.daemon = True
            scroll_thread.start()
            return
        else:
            # Split and display in 2 lines if it fits
            line1 = message[:16]
            line2 = message[16:32] if len(message) > 16 else ""
            setText(f"{line1}\n{line2}")
        
    except Exception as e:
        log_event("ERROR", f"Couldnt update LCD: {e}")


def scroll_text(message, delay=0.3, loop_count=1, idle_message="Smart Doorbell Ready"):
    message = message.strip() + "    "  # Padding
    display_width = 16

    try:
        for _ in range(loop_count):
            for i in range(len(message) - display_width + 1):
                line1 = message[i:i + display_width]
                setText(f"{line1}\n")  # Second line blank
                time.sleep(delay)

        set_display(idle_message)

    except Exception as e:
        print(f"[LCD Scroll Error] {e}")

def adjust_lights():
    global night_mode
    night_mode = not night_mode
    log_event("EVENT", f"Night mode set to {night_mode}")


# === System Functions ===
def log_event(mtype, message):
    timestamp = datetime.datetime.now().strftime('%d-%m-%Y %H:%M:%S')
    print(f"[{mtype}] ({timestamp}): {message}")
    return

def self_check():
    status = {
        "PIR": False,
        "Light": False,
        "Button": False,
        "LCD": False
    }

    # Check LCD
    try:
        set_display("Running self check...")
        time.sleep(1)
        log_event("PASS", "LCD ON")
        status["LCD"] = True
    except:
        log_event("ERROR", "LCD not responding.")

    # Check PIR
    try:
        motion = digitalRead(pir_port)
        log_event("PASS", f"PIR motion sensor: {motion}")
        status["PIR"] = True
    except:
        log_event("ERROR", "PIR motion sensor not responding")

    # Check Light Level
    try:
        light = analogRead(light_port)
        if 0 <= light <= 1023:
            log_event("PASS", f"Light level sensor: {light}")
            status["Light"] = True
        else:
            log_event("ERROR", "Light sensor value out of range.")
    except:
        log_event("ERROR", "Light sensor not responding.")

    # Check Button
    try:
        state = digitalRead(button_port)
        log_event("PASS", f"Button state: {'Pressed' if state == 0 else 'Released'}")
        status["Button"] = True
    except:
        log_event("ERROR", "Button not responding.")

    # Final Summary
    failed = [key for key, val in status.items() if not val]
    if not failed:
        set_display("All systems operational")
        log_event("SUCCESS", "All sensors passed.")
        return True
    else:
        set_display("SENSORS FAILED: ".join(failed))
        log_event("FAILED", f"Sensors failed: {', '.join(failed)}")
        return False

def send_data_thingsboard(payload):
    global client
    if not client:
        log_event("ERROR","MQTT Client not initilised")
        return
    
    try:
        client.publish("v1/devices/me/telemetry", json.dumps(payload), qos=1)
        log_event("THINGSBOARD",f"Data sent: {payload}")
    except Exception as e:
        log_event("ERROR",f"failed to send data to thingsboard: {e}")
     
def handle_attribute_update(client, userdata, msg):
    global preferences
    try:
        data = json.loads(msg.payload.decode("utf-8"))
        updated = []
        for key in preferences:
            if key in data:
                preferences[key] = data[key]
                updated.append(f"{key} = {data[key]}")
        if updated:
            log_event("THINGSBOARD", "Preferences Updated: " + ", ".join(updated))
    except Exception as e:
        log_event("ERROR",f"Failed to handle attribute update: {e}")
        
def report_preferences():
    try:
        client.publish("v1/devices/me/attributes", json.dumps(preferences), qos =1)
        log_event("THINGSBOARD","Reported current preferences")
    except Exception as e:
        log_event("ERROR","Failed to report preferences: {e}")
        
# === Main Reaction Loop ===
def main_loop():
    print("System Starting")
    set_display("Smart Doorbell Ready")

    try:
        while True:
            detect_button_press()
            detect_motion()
            detect_light_level()
            #print("[DEBUG]: LOOP")
            time.sleep(0.2)

    except KeyboardInterrupt:
        print("Smart doorbell shutdown.")
        set_display("SYSTEM HALTED KEYBOARD INTERRUPT")

# Run
if __name__ == "__main__":
    initialise_sensors()
    initilise_thingsboard()
    report_preferences()
    self_check()
    main_loop()
